// BitmapFont.cs
// Bitmap Font class for XNA
// Copyright 2006 Microsoft Corp.
// Gary Kacmarcik (garykac@microsoft.com)
// Revision: 2006-Nov-18

using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;

namespace XNAExtras
{
    public enum TextAlignment
    {
        Left,
        Center,
        Right,
    }

    public enum GlyphFlags
    {
        None = 0,
        ForceWhite = 1,		// force the drawing color for this glyph to be white.
    }

    /// <summary>
    /// Info for each glyph in the font - where to find the glyph image and 
    /// other properties
    /// </summary>
    public struct GlyphInfo
    {
        public ushort nBitmapID;
        public byte pxLocX;
        public byte pxLocY;
        public byte pxWidth;
        public byte pxHeight;
        public byte pxAdvanceWidth;
        public sbyte pxLeftSideBearing;
        public GlyphFlags nFlags;
    }

    /// <summary>
	/// Bitmap font class for XNA
	/// </summary>
    public partial class BitmapFont : IDisposable
    {
        /// <summary>
        /// Save font rendering state so that it can be saved/restored
        /// </summary>
        public struct SaveStateInfo
        {
            public bool fKern;
            public float fpDepth;
            public TextAlignment align;
            public Vector2 vPen;
            public Color color;
        }

        private SpriteBatch m_sb;
        private SpriteBatch m_sbOverride;
        private string m_strName;
        private Dictionary<int, Texture2D> m_dictBitmapID2Texture;
        private Dictionary<char, GlyphInfo> m_dictUnicode2GlyphInfo;
        private Dictionary<char, Dictionary<char, sbyte>> m_dictKern;
        private int m_nBase = 0;
        private int m_nHeight = 0;
        private float m_fpDepth = 0.0f;
        private TextAlignment m_eAlign = TextAlignment.Left;

        /// <summary>
        /// Create an empty bitmap font to be filled in by a BitmapFontReader.
        /// Use content.Load<BitmapFont>(fontname);
        /// </summary>
        public BitmapFont(GraphicsDevice device)
        {
            m_sb = new SpriteBatch(device);
            m_dictBitmapID2Texture = new Dictionary<int, Texture2D>();
            m_dictUnicode2GlyphInfo = new Dictionary<char, GlyphInfo>();
            m_dictKern = new Dictionary<char, Dictionary<char, sbyte>>();
        }

        /// <summary>
        /// Destructor for BitmapFont.
        /// </summary>
        ~BitmapFont()
        {
            Dispose();
        }

        /// <summary>
        /// Dispose of all of the non-managed resources for this object
        /// </summary>
        public void Dispose()
        {
            m_sb.Dispose();
            foreach (KeyValuePair<int, Texture2D> pair in m_dictBitmapID2Texture)
                pair.Value.Dispose();
        }

        /// <summary>
        /// The name of this font
        /// </summary>
        public string Name
        {
            get { return m_strName; }
        }

        /// <summary>
        /// Should we kern adjacent characters?
        /// </summary>
        private bool m_fKern = true;

        /// <summary>
        /// Enable/disable kerning
        /// </summary>
        public bool KernEnable
        {
            get { return m_fKern; }
            set { m_fKern = value; }
        }

        /// <summary>
        /// Distance from top of font to the baseline
        /// </summary>
        public int Baseline
        {
            get { return m_nBase; }
        }

        /// <summary>
        /// Distance from top to bottom of the font
        /// </summary>
        public int LineHeight
        {
            get { return m_nHeight; }
        }

        /// <summary>
        /// The depth at which to draw the font
        /// </summary>
        public float Depth
        {
            get { return m_fpDepth; }
            set { m_fpDepth = value; }
        }

        /// <summary>
        /// The text alignment. This is only used by the TextBox routines
        /// </summary>
        public TextAlignment Alignment
        {
            get { return m_eAlign; }
            set { m_eAlign = value; }
        }

        /// <summary>
        /// Calculate the width of the given string.
        /// </summary>
        /// <param name="format">String format</param>
        /// <param name="args">String format arguments</param>
        /// <returns>Width (in pixels) of the string</returns>
        public int MeasureString(string format, params object[] args)
        {
            string str = string.Format(format, args);
            int pxWidth = 0;
            char cLast = '\0';

            foreach (char c in str)
            {
                if (!m_dictUnicode2GlyphInfo.ContainsKey(c))
                {
                    //TODO: print out undefined char glyph
                    continue;
                }

                GlyphInfo ginfo = m_dictUnicode2GlyphInfo[c];

                // if kerning is enabled, get the kern adjustment for this char pair
                if (m_fKern)
                {
                    pxWidth += CalcKern(cLast, c);
                    cLast = c;
                }

                // update the string width
                pxWidth += ginfo.pxAdvanceWidth;
            }

            return pxWidth;
        }

        /// <summary>
        /// Calculate the number of characters that fit in the given width.
        /// </summary>
        /// <param name="pxMaxWidth">Maximum string width</param>
        /// <param name="str">String</param>
        /// <param name="nChars">Number of characters that fit</param>
        /// <param name="pxWidth">Width of substring</param>
        private void CountCharWidth(int pxMaxWidth, string str, out int nChars, out int pxWidth)
        {
            int nLastWordBreak = 0;
            int pxLastWordBreakWidth = 0;
            int pxLastWidth = 0;
            char cLast = '\0';

            nChars = 0;
            pxWidth = 0;

            foreach (char c in str)
            {
                // if this is a newline, then return. the width is set correctly
                if (c == '\n')
                {
                    nChars++;
                    return;
                }

                if (!m_dictUnicode2GlyphInfo.ContainsKey(c))
                {
                    //TODO: print out undefined char glyph
                    continue;
                }

                GlyphInfo ginfo = m_dictUnicode2GlyphInfo[c];

                // if kerning is enabled, get the kern adjustment for this char pair
                if (m_fKern)
                {
                    int pxKern = CalcKern(cLast, c);
                    pxWidth += pxKern;
                    cLast = c;
                }

                // update the string width and char count
                pxLastWidth = pxWidth;
                pxWidth += ginfo.pxAdvanceWidth;
                nChars++;

                // record the end of the previous word if this is a whitespace char
                if (Char.IsWhiteSpace(c))
                {
                    nLastWordBreak = nChars;			// include space in char count
                    pxLastWordBreakWidth = pxLastWidth;	// don't include space in width
                }

                // if we've exceeded the max, then return the chars up to the last complete word
                if (pxWidth > pxMaxWidth)
                {
                    pxWidth = pxLastWordBreakWidth;
                    if (pxWidth == 0)
                    {
                        // fallback to last char if we haven't seen a complete word
                        pxWidth = pxLastWidth;
                        nChars--;
                    }
                    else
                        nChars = nLastWordBreak;
                    return;
                }
            }
        }

        /// <summary>
        /// Current pen position
        /// </summary>
        private Vector2 m_vPen = new Vector2(0, 0);

        /// <summary>
        /// Current pen position
        /// </summary>
        public Vector2 Pen
        {
            get { return m_vPen; }
            set { m_vPen = value; }
        }

        /// <summary>
        /// Set the current pen position
        /// </summary>
        /// <param name="x">X-coord</param>
        /// <param name="y">Y-coord</param>
        public void SetPen(int x, int y)
        {
            m_vPen = new Vector2(x, y);
        }


        /// <summary>
        /// Current color used for drawing text
        /// </summary>
        private Color m_color = Color.White;

        /// <summary>
        /// Current color used for drawing text
        /// </summary>
        public Color TextColor
        {
            get { return m_color; }
            set { m_color = value; }
        }


        /// <summary>
        /// Draw the given string at (x,y).
        /// The text color is inherited from the last draw command (default=White).
        /// </summary>
        /// <param name="x">X-coord</param>
        /// <param name="y">Y-coord</param>
        /// <param name="format">String format</param>
        /// <param name="args">String format args</param>
        /// <returns>Width of string (in pixels)</returns>
        public int DrawString(int x, int y, string format, params object[] args)
        {
            Vector2 v = new Vector2(x, y);
            return DrawString(v, m_color, format, args);
        }

        /// <summary>
        /// Draw the given string at (x,y) using the specified color
        /// </summary>
        /// <param name="x">X-coord</param>
        /// <param name="y">Y-coord</param>
        /// <param name="color">Text color</param>
        /// <param name="format">String format</param>
        /// <param name="args">String format args</param>
        /// <returns>Width of string (in pixels)</returns>
        public int DrawString(int x, int y, Color color, string format, params object[] args)
        {
            Vector2 v = new Vector2(x, y);
            return DrawString(v, color, format, args);
        }

        /// <summary>
        /// Draw the given string using the specified color.
        /// The text drawing location is immediately after the last drawn text (default=0,0).
        /// </summary>
        /// <param name="color">Text color</param>
        /// <param name="format">String format</param>
        /// <param name="args">String format args</param>
        /// <returns>Width of string (in pixels)</returns>
        public int DrawString(Color color, string format, params object[] args)
        {
            return DrawString(m_vPen, color, format, args);
        }

        /// <summary>
        /// Draw the given string at (x,y).
        /// The text drawing location is immediately after the last drawn text (default=0,0).
        /// The text color is inherited from the last draw command (default=White).
        /// </summary>
        /// <param name="format">String format</param>
        /// <param name="args">String format args</param>
        /// <returns>Width of string (in pixels)</returns>
        public int DrawString(string format, params object[] args)
        {
            return DrawString(m_vPen, m_color, format, args);
        }

        /// <summary>
        /// Draw the given string at vOrigin using the specified color
        /// </summary>
        /// <param name="vAt">(x,y) coord</param>
        /// <param name="cText">Text color</param>
        /// <param name="strFormat">String format</param>
        /// <param name="args">String format args</param>
        /// <returns>Width of string (in pixels)</returns>
        public int DrawString(Vector2 vAt, Color cText, string strFormat, params object[] args)
        {
            string str = string.Format(strFormat, args);

            return DrawString_internal(vAt, cText, str);
        }

        /// <summary>
        /// Private version of DrawString that expects the string to be formatted already
        /// </summary>
        /// <param name="vAt">(x,y) coord</param>
        /// <param name="cText">Text color</param>
        /// <param name="str">String</param>
        /// <returns>Width of string (in pixels)</returns>
        private int DrawString_internal(Vector2 vAt, Color cText, string str)
        {
            Vector2 vOrigin = new Vector2(0, 0);
            int pxWidth = 0;
            char cLast = '\0';

            // are we using our local SpriteBatch, or an override?
            bool fSBOverride = (m_sbOverride != null);
            SpriteBatch sb = (fSBOverride ? m_sbOverride : m_sb);

            if (!fSBOverride)
                sb.Begin(SpriteBlendMode.AlphaBlend);

            // draw each character in the string
            foreach (char c in str)
            {
                if (!m_dictUnicode2GlyphInfo.ContainsKey(c))
                {
                    //TODO: print out undefined char glyph
                    continue;
                }

                GlyphInfo ginfo = m_dictUnicode2GlyphInfo[c];

                // if kerning is enabled, get the kern adjustment for this char pair
                if (m_fKern)
                {
                    int pxKern = CalcKern(cLast, c);
                    vAt.X += pxKern;
                    pxWidth += pxKern;
                    cLast = c;
                }

                // draw the glyph
                vAt.X += ginfo.pxLeftSideBearing;
                if (ginfo.pxWidth != 0 && ginfo.pxHeight != 0)
                {
                    Rectangle rSource = new Rectangle(ginfo.pxLocX, ginfo.pxLocY, ginfo.pxWidth, ginfo.pxHeight);
                    Color color = (((ginfo.nFlags & GlyphFlags.ForceWhite) != 0) ? Color.White : cText);
                    sb.Draw(m_dictBitmapID2Texture[ginfo.nBitmapID], vAt, rSource, color, 0.0f, vOrigin, 1.0f, SpriteEffects.None, m_fpDepth);
                }

                // update the string width and advance the pen to the next drawing position
                pxWidth += ginfo.pxAdvanceWidth;
                vAt.X += ginfo.pxAdvanceWidth - ginfo.pxLeftSideBearing;
            }

            if (!fSBOverride)
                sb.End();

            // record final pen position and color
            m_vPen = vAt;
            m_color = cText;

            return pxWidth;
        }

        /// <summary>
        /// Get the kern value for the given pair of characters
        /// </summary>
        /// <param name="chLeft">Left character</param>
        /// <param name="chRight">Right character</param>
        /// <returns>Amount to kern (in pixels)</returns>
        private int CalcKern(char chLeft, char chRight)
        {
            if (m_dictKern.ContainsKey(chLeft))
            {
                Dictionary<char, sbyte> kern2 = m_dictKern[chLeft];
                if (kern2.ContainsKey(chRight))
                    return kern2[chRight];
            }
            return 0;
        }

        /// <summary>
        /// Draw text formatted to fit in the specified rectangle
        /// </summary>
        /// <param name="r">The rectangle to fit the text</param>
        /// <param name="cText">Text color</param>
        /// <param name="strFormat">String format</param>
        /// <param name="args">String format args</param>
        public void TextBox(Rectangle r, Color cText, string strFormat, params object[] args)
        {
            string str = string.Format(strFormat, args);

            int nChars;
            int pxWidth;
            Vector2 vAt = new Vector2(r.Left, r.Top);

            while (str.Length != 0)
            {
                // stop drawing if there isn't room for this line
                if (vAt.Y + m_nHeight > r.Bottom)
                    return;

                CountCharWidth(r.Width, str, out nChars, out pxWidth);

                switch (m_eAlign)
                {
                    case TextAlignment.Left:
                        vAt.X = r.Left;
                        break;
                    case TextAlignment.Center:
                        vAt.X = r.Left + ((r.Width - pxWidth) / 2);
                        break;
                    case TextAlignment.Right:
                        vAt.X = r.Left + (r.Width - pxWidth);
                        break;
                }
                DrawString_internal(vAt, cText, str.Substring(0, nChars));
                str = str.Substring(nChars);
                vAt.Y += m_nHeight;
            }
        }

        /// <summary>
        /// Save the current font rendering state
        /// </summary>
        /// <param name="bfss">Struct to store the save state</param>
        public void SaveState(out SaveStateInfo bfss)
        {
            bfss.fKern = m_fKern;
            bfss.fpDepth = m_fpDepth;
            bfss.align = m_eAlign;
            bfss.vPen = m_vPen;
            bfss.color = m_color;
        }

        /// <summary>
        /// Restore the font rendering state
        /// </summary>
        /// <param name="bfss">Previously saved font state</param>
        public void RestoreState(SaveStateInfo bfss)
        {
            m_fKern = bfss.fKern;
            m_fpDepth = bfss.fpDepth;
            m_eAlign = bfss.align;
            m_vPen = bfss.vPen;
            m_color = bfss.color;
        }

        /// <summary>
        /// Temporarily override the font's SpriteBatch with the given SpriteBatch.
        /// </summary>
        /// <param name="sb">The new SpriteBatch (or null to reset)</param>
        /// <remarks>
        /// When drawing text using the SpriteBatch override, Begin/End will not be called on the SpriteBatch.
        /// Use null to reset back to the font's original SpriteBatch.
        /// </remarks>
        public void SpriteBatchOverride(SpriteBatch sb)
        {
            m_sbOverride = sb;
        }

        /// <summary>
        /// Loads the font via the ContentPipeline.  Must match format exported by
        /// XNAImport.BitmapFontData's Write method.
        /// </summary>
        /// <param name="input">ContentReader instance from our specific ContentTypeReader</param>
        public void Read(ContentReader input)
        {
            int count = input.ReadInt32();
            for (int i = 0; i < count; i++)
            {
                int key = input.ReadInt32();
                m_dictBitmapID2Texture[key] = input.ReadExternalReference<Texture2D>();
            }

            count = input.ReadInt32();
            for (int i = 0; i < count; i++)
            {
                char key = input.ReadChar();
                GlyphInfo info;
                info.nBitmapID = input.ReadUInt16();
                info.pxLocX = input.ReadByte();
                info.pxLocY = input.ReadByte();
                info.pxWidth = input.ReadByte();
                info.pxHeight = input.ReadByte();
                info.pxAdvanceWidth = input.ReadByte();
                info.pxLeftSideBearing = input.ReadSByte();
                info.nFlags = (GlyphFlags)input.ReadByte();
                m_dictUnicode2GlyphInfo[key] = info;
            }

            count = input.ReadInt32();
            for (int i = 0; i < count; i++)
            {
                char key = input.ReadChar();
                Dictionary<char, sbyte> info = new Dictionary<char, sbyte>();
                int num = input.ReadInt32();
                for (int n = 0; n < num; n++)
                {
                    char c = input.ReadChar();
                    info[c] = input.ReadSByte();
                }
                m_dictKern[key] = info;
            }

            m_strName = input.ReadString();
            m_nBase = input.ReadInt32();
            m_nHeight = input.ReadInt32();
            m_fpDepth = input.ReadSingle();
            m_eAlign = (TextAlignment)input.ReadByte();
        }
    }

    /// <summary>
    /// ContentPipeline reader for BitmapFont.
    /// </summary>
    class BitmapFontReader : ContentTypeReader<BitmapFont>
    {
        protected override BitmapFont Read(ContentReader input, BitmapFont existingInstance)
        {
            ContentManager content = input.ContentManager;
            IGraphicsDeviceService service = (IGraphicsDeviceService)
                content.ServiceProvider.GetService(typeof(IGraphicsDeviceService));
            GraphicsDevice device = service.GraphicsDevice;

            BitmapFont font = new BitmapFont(device);
            font.Read(input);
            return font;
        }
    }
}
