using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Content.Pipeline;
using Microsoft.Xna.Framework.Content.Pipeline.Graphics;
using Microsoft.Xna.Framework.Content.Pipeline.Serialization.Compiler;
using System.Text;
using System.Xml;
using System.IO;

namespace XNAExtras
{
    class BitmapFontDescription
    {
        public BitmapFontDescription(string name, string path, XmlNodeList xnl)
        {
            this.name = name;
            this.path = path;
            this.xnl = xnl;
        }

        private string name;

        public string Name
        {
            get { return name; }
        }

        private string path;

        public string Path
        {
            get { return path; }
        }

        private XmlNodeList xnl;

        public XmlNodeList Xnl
        {
            get { return xnl; }
        }
    }

    [ContentImporter(".xml", DefaultProcessor = "BitmapFontProcessor")]
    class BitmapFontImporter : ContentImporter<BitmapFontDescription>
    {
        public override BitmapFontDescription Import(string filename, ContentImporterContext context)
        {
            string name = Path.GetFileNameWithoutExtension(filename);
            string path = Path.GetDirectoryName(filename);

            XmlDocument xd = new XmlDocument();
            xd.Load(filename);

            return new BitmapFontDescription(name, path, xd.ChildNodes);
        }
    }

    partial class BitmapFontData
    {
        private Dictionary<int, ExternalReference<TextureContent>> m_dictBitmapID2Texture;
        private Dictionary<char, GlyphInfo> m_dictUnicode2GlyphInfo;
        private Dictionary<char, Dictionary<char, sbyte>> m_dictKern;
        private string m_strName;
        private string m_strPath;
        private int m_nBase = 0;
        private int m_nHeight = 0;
        private float m_fpDepth = 0.0f;
        private TextAlignment m_eAlign = TextAlignment.Left;

        public BitmapFontData(BitmapFontDescription input, ContentProcessorContext context)
        {
            m_dictBitmapID2Texture = new Dictionary<int, ExternalReference<TextureContent>>();
            m_dictUnicode2GlyphInfo = new Dictionary<char, GlyphInfo>();
            m_dictKern = new Dictionary<char, Dictionary<char, sbyte>>();

            m_strName = input.Name;
            m_strPath = input.Path;
            if (m_strPath != "")
                m_strPath += @"\";

            LoadFontXML(input.Xnl, context);
        }

        /// <summary>
        /// Load the font data from an XML font descriptor file
        /// </summary>
        /// <param name="xnl">XML node list containing the entire font descriptor file</param>
        private void LoadFontXML(XmlNodeList xnl, ContentProcessorContext context)
        {
            foreach (XmlNode xn in xnl)
            {
                if (xn.Name == "font")
                {
                    m_strName = GetXMLAttribute(xn, "name");
                    m_nBase = Int32.Parse(GetXMLAttribute(xn, "base"));
                    m_nHeight = Int32.Parse(GetXMLAttribute(xn, "height"));

                    LoadFontXML_font(xn.ChildNodes, context);
                }
            }
        }

        /// <summary>
        /// Load the data from the "font" node
        /// </summary>
        /// <param name="xnl">XML node list containing the "font" node's children</param>
        private void LoadFontXML_font(XmlNodeList xnl, ContentProcessorContext context)
        {
            foreach (XmlNode xn in xnl)
            {
                if (xn.Name == "bitmaps")
                    LoadFontXML_bitmaps(xn.ChildNodes, context);
                if (xn.Name == "glyphs")
                    LoadFontXML_glyphs(xn.ChildNodes, context);
                if (xn.Name == "kernpairs")
                    LoadFontXML_kernpairs(xn.ChildNodes, context);
            }
        }

        /// <summary>
        /// Load the data from the "bitmaps" node
        /// </summary>
        /// <param name="xnl">XML node list containing the "bitmaps" node's children</param>
        private void LoadFontXML_bitmaps(XmlNodeList xnl, ContentProcessorContext context)
        {
            foreach (XmlNode xn in xnl)
            {
                if (xn.Name == "bitmap")
                {
                    string strID = GetXMLAttribute(xn, "id");
                    string strFilename = GetXMLAttribute(xn, "name");

                    m_dictBitmapID2Texture[Int32.Parse(strID)] =
                        context.BuildAsset<Texture2DContent, TextureContent>(
                            new ExternalReference<Texture2DContent>(m_strPath + strFilename),
                            "SpriteTextureProcessor");
                }
            }
        }

        /// <summary>
        /// Load the data from the "glyphs" node
        /// </summary>
        /// <param name="xnl">XML node list containing the "glyphs" node's children</param>
        private void LoadFontXML_glyphs(XmlNodeList xnl, ContentProcessorContext context)
        {
            foreach (XmlNode xn in xnl)
            {
                if (xn.Name == "glyph")
                {
                    string strChar = GetXMLAttribute(xn, "ch");
                    string strBitmapID = GetXMLAttribute(xn, "bm");
                    string strLoc = GetXMLAttribute(xn, "loc");
                    string strSize = GetXMLAttribute(xn, "size");
                    string strAW = GetXMLAttribute(xn, "aw");
                    string strLSB = GetXMLAttribute(xn, "lsb");
                    string strForceWhite = GetXMLAttribute(xn, "forcewhite");

                    if (strLoc == "")
                        strLoc = GetXMLAttribute(xn, "origin");	// obsolete - use loc instead

                    string[] aLoc = strLoc.Split(',');
                    string[] aSize = strSize.Split('x');

                    GlyphInfo ginfo = new GlyphInfo();
                    ginfo.nBitmapID = UInt16.Parse(strBitmapID);
                    ginfo.pxLocX = Byte.Parse(aLoc[0]);
                    ginfo.pxLocY = Byte.Parse(aLoc[1]);
                    ginfo.pxWidth = Byte.Parse(aSize[0]);
                    ginfo.pxHeight = Byte.Parse(aSize[1]);
                    ginfo.pxAdvanceWidth = Byte.Parse(strAW);
                    ginfo.pxLeftSideBearing = SByte.Parse(strLSB);
                    ginfo.nFlags = 0;
                    ginfo.nFlags |= (strForceWhite == "true" ? GlyphFlags.ForceWhite : GlyphFlags.None);

                    m_dictUnicode2GlyphInfo[strChar[0]] = ginfo;
                }
            }
        }

        /// <summary>
        /// Load the data from the "kernpairs" node
        /// </summary>
        /// <param name="xnl">XML node list containing the "kernpairs" node's children</param>
        private void LoadFontXML_kernpairs(XmlNodeList xnl, ContentProcessorContext context)
        {
            foreach (XmlNode xn in xnl)
            {
                if (xn.Name == "kernpair")
                {
                    string strLeft = GetXMLAttribute(xn, "left");
                    string strRight = GetXMLAttribute(xn, "right");
                    string strAdjust = GetXMLAttribute(xn, "adjust");

                    char chLeft = strLeft[0];
                    char chRight = strRight[0];

                    // create a kern dict for the left char if needed
                    if (!m_dictKern.ContainsKey(chLeft))
                        m_dictKern[chLeft] = new Dictionary<char, sbyte>();

                    // add the right char to the left char's kern dict
                    Dictionary<char, sbyte> kern2 = m_dictKern[chLeft];
                    kern2[chRight] = SByte.Parse(strAdjust);
                }
            }
        }

        /// <summary>
        /// Get the XML attribute value
        /// </summary>
        /// <param name="n">XML node</param>
        /// <param name="strAttr">Attribute name</param>
        /// <returns>Attribute value, or the empty string if the attribute doesn't exist</returns>
        private static string GetXMLAttribute(XmlNode n, string strAttr)
        {
            XmlAttribute attr = n.Attributes.GetNamedItem(strAttr) as XmlAttribute;
            if (attr != null)
                return attr.Value;
            return "";
        }
    }

    [ContentProcessor]
    class BitmapFontProcessor : ContentProcessor<BitmapFontDescription, BitmapFontData>
    {
        public override BitmapFontData Process(BitmapFontDescription input, ContentProcessorContext context)
        {
            return new BitmapFontData(input, context);
        }
    }

    [ContentTypeWriter]
    class BitmapFontWriter : ContentTypeWriter<BitmapFontData>
    {
        protected override void Write(ContentWriter output, BitmapFontData value)
        {
            value.Write(output);
        }

        public override string GetRuntimeType(TargetPlatform targetPlatform)
        {
            return typeof(BitmapFont).AssemblyQualifiedName;
        }

        public override string GetRuntimeReader(TargetPlatform targetPlatform)
        {
            return "XNAExtras.BitmapFontReader, XNABase, Version=1.0.0.0, Culture=neutral";
        }
    }

    partial class BitmapFontData
    {
        public void Write(ContentWriter output)
        {
            output.Write(m_dictBitmapID2Texture.Count);
            foreach (KeyValuePair<int, ExternalReference<TextureContent>> pair in m_dictBitmapID2Texture)
            {
                output.Write(pair.Key);
                output.WriteExternalReference(pair.Value);
            }

            output.Write(m_dictUnicode2GlyphInfo.Count);
            foreach (KeyValuePair<char, GlyphInfo> pair in m_dictUnicode2GlyphInfo)
            {
                output.Write(pair.Key);
                output.Write(pair.Value.nBitmapID);
                output.Write(pair.Value.pxLocX);
                output.Write(pair.Value.pxLocY);
                output.Write(pair.Value.pxWidth);
                output.Write(pair.Value.pxHeight);
                output.Write(pair.Value.pxAdvanceWidth);
                output.Write(pair.Value.pxLeftSideBearing);
                output.Write((byte)pair.Value.nFlags);
            }

            output.Write(m_dictKern.Count);
            foreach (KeyValuePair<char, Dictionary<char, sbyte>> pair in m_dictKern)
            {
                output.Write(pair.Key);
                output.Write(pair.Value.Count);
                foreach (KeyValuePair<char, sbyte> p in pair.Value)
                {
                    output.Write(p.Key);
                    output.Write(p.Value);
                }
            }

            output.Write(m_strName);
            output.Write(m_nBase);
            output.Write(m_nHeight);
            output.Write(m_fpDepth);
            output.Write((byte)m_eAlign);
        }
    }
}
