#region File Description
//-----------------------------------------------------------------------------
// File:      Bar.cs
// Namespace: WindowSystem
// Author:    Aaron MacDougall
//-----------------------------------------------------------------------------
#endregion

#region Using Statements
using System.Diagnostics;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
#endregion

namespace XNAWindowSystem
{
    /// <summary>
    /// A graphical bar drawn with two end sprites, and a stretched middle
    /// sprite.
    /// </summary>
    /// <remarks>
    /// Inherited from SkinnedComponent to allow arbitrary number of skin
    /// states. For example, the Scrollbar class uses a Bar to draw the thumb
    /// with three states, normal, hover, and pressed.
    /// </remarks>
    public class Bar : SkinnedComponent
    {
        #region Default Properties
        private static int defaultEdgeSize = 5;

        /// <summary>
        /// Sets the default size of the edge of the bar skin.
        /// </summary>
        /// <value>Must be at least 0.</value>
        public static int DefaultEdgeSize
        {
            set
            {
                Debug.Assert(value >= 0);
                defaultEdgeSize = value;
            }
        }
        #endregion

        #region Fields
        private int edgeSize;
        private bool isVertical;
        #endregion

        #region Properties
        /// <summary>
        /// Get/Set whether the bar is vertical or horizontal.
        /// </summary>
        /// <value>TRUE if bar is vertical, otherwise FALSE.</value>
        public bool IsVertical
        {
            get { return isVertical; }
            set
            {
                this.isVertical = value;
                RefreshSkins();
            }
        }

        /// <summary>
        /// Get/Set the size of the edge of the bar skin.
        /// </summary>
        /// <value>Must be at least 0.</value>
        public int EdgeSize
        {
            get { return edgeSize; }
            set
            {
                Debug.Assert(value >= 0);

                this.edgeSize = value;
                RefreshSkins();
            }
        }
        #endregion

        #region Constructors
        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="game">The currently running Game object.</param>
        /// <param name="guiManager">GUIManager that this control is part of.</param>
        public Bar(Game game, GUIManager guiManager)
            : base(game, guiManager)
        {
            this.isVertical = false;
            CanHaveFocus = false;

            #region Set Default Properties
            EdgeSize = defaultEdgeSize;
            #endregion
        }
        #endregion

        /// <summary>
        /// Update skin sizes.
        /// </summary>
        protected override void RefreshSkins()
        {
            // Get the current list of skins
            Dictionary<int, ComponentSkin> skins = this.Skins;

            foreach (KeyValuePair<int, ComponentSkin> skin in skins)
            {
                skin.Value.Rects = CreateBar(
                    GetSkinLocation(skin.Key),
                    new Rectangle(0, 0, Width, Height),
                    this.edgeSize,
                    this.isVertical
                );
            }

            Redraw();
        }

        /// <summary>
        /// Constructs a bar skin by creating edges and stretching the middle.
        /// </summary>
        /// <param name="source">Location on source texture.</param>
        /// <param name="dimensions">Size of bar to construct.</param>
        /// <param name="edgeSize">The bar edge size.</param>
        /// <param name="isVertical">Orientation of the bar.</param>
        /// <returns>List of skin locations of new bar.</returns>
        protected static List<GUIRect> CreateBar(
            Rectangle source,
            Rectangle dimensions,
            int edgeSize,
            bool isVertical
            )
        {
            List<GUIRect> result = new List<GUIRect>();
            GUIRect[] sprites = new GUIRect[3];

            if (isVertical)
            {
                // Top
                sprites[0] = new GUIRect();
                sprites[0].Source = new Rectangle(source.X, source.Y, source.Width, edgeSize);
                sprites[0].Destination = new Rectangle(dimensions.X, dimensions.Y, dimensions.Width, edgeSize);

                // Middle
                sprites[1] = new GUIRect();
                sprites[1].Source = new Rectangle(source.X, source.Y + edgeSize, source.Width, source.Height - (2 * edgeSize));
                sprites[1].Destination = new Rectangle(dimensions.X, dimensions.Y + edgeSize, dimensions.Width, dimensions.Height - (2 * edgeSize));

                // Bottom
                sprites[2] = new GUIRect();
                sprites[2].Source = new Rectangle(source.X, source.Bottom - edgeSize, source.Width, edgeSize);
                sprites[2].Destination = new Rectangle(dimensions.X, dimensions.Y + dimensions.Height - edgeSize, dimensions.Width, edgeSize);
            }
            else
            {
                // Left
                sprites[0] = new GUIRect();
                sprites[0].Source = new Rectangle(source.X, source.Y, edgeSize, source.Height);
                sprites[0].Destination = new Rectangle(dimensions.X, dimensions.Y, edgeSize, dimensions.Height);

                // Middle
                sprites[1] = new GUIRect();
                sprites[1].Source = new Rectangle(source.X + edgeSize, source.Y, source.Width - (2 * edgeSize), source.Height);
                sprites[1].Destination = new Rectangle(dimensions.X + edgeSize, dimensions.Y, dimensions.Width - (2 * edgeSize), dimensions.Height);

                // Right
                sprites[2] = new GUIRect();
                sprites[2].Source = new Rectangle(source.Right - edgeSize, source.Y, edgeSize, source.Height);
                sprites[2].Destination = new Rectangle(dimensions.Right - edgeSize, dimensions.Y, edgeSize, dimensions.Height);
            }

            // Copy over the results
            for (int i = 0; i < 3; i++)
                result.Add(sprites[i]);

            return result;
        }
    }
}