#region File Description
//-----------------------------------------------------------------------------
// File:      CheckBox.cs
// Namespace: WindowSystem
// Author:    Aaron MacDougall
//-----------------------------------------------------------------------------
#endregion

#region Using Statements
using System;
using System.Diagnostics;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using XNAExtras;
#endregion

namespace XNAWindowSystem
{
    /// <summary>
    /// A graphical checkbox button, with a description label.
    /// </summary>
    public class CheckBox : DrawableUIComponent
    {
        #region Default Properties
        private static int defaultHeight = 15;
        private static int defaultHMargin = 5;
        private static string defaultFont = "Content/Fonts/DefaultFont";
        private static DefaultSixSkins defaultButtonSkin = new DefaultSixSkins(
            new Rectangle(1, 27, 15, 15),
            new Rectangle(17, 27, 15, 15),
            new Rectangle(33, 27, 15, 15),
            new Rectangle(1, 43, 15, 15),
            new Rectangle(17, 43, 15, 15),
            new Rectangle(33, 43, 15, 15)
            );

        /// <summary>
        /// Sets the default control height, which is also the width and height
        /// of the button.
        /// </summary>
        /// <value>Must be greater than 0.</value>
        public static int DefaultHeight
        {
            set
            {
                Debug.Assert(value > 0);
                defaultHeight = value;
            }
        }

        /// <summary>
        /// Sets the default gap between the button and the label.
        /// </summary>
        /// <value>Must be at least 0.</value>
        public static int DefaultHMargin
        {
            set
            {
                Debug.Assert(value >= 0);
                defaultHMargin = value;
            }
        }

        /// <summary>
        /// Sets the default font for CheckBox labels.
        /// </summary>
        /// <value>Must not be null.</value>
        public static string DefaultFont
        {
            set
            {
                Debug.Assert(value != null);
                defaultFont = value;
            }
        }

        /// <summary>
        /// Gets the default skin locations.
        /// </summary>
        public static ISixSkins DefaultButtonSkin
        {
            get { return defaultButtonSkin; }
        }
        #endregion

        #region Fields
        private ImageButton button;
        private Label label;
        private int hMargin;
        #endregion

        #region Properties
        /// <summary>
        /// Sets the gap between the button and the label.
        /// </summary>
        /// <value>Must be at least 0.</value>
        public int HMargin
        {
            get { return hMargin; }
            set
            {
                Debug.Assert(value >= 0);

                hMargin = value;
                RefreshMargins();
            }
        }

        /// <summary>
        /// Sets the label font.
        /// </summary>
        /// <value>Must not be null.</value>
        public BitmapFont Font
        {
            set { label.Font = value; }
        }

        /// <summary>
        /// Get/Set the description text.
        /// </summary>
        /// <value>Must not be null.</value>
        public string Text
        {
            get { return label.Text; }
            set
            {
                label.Text = value;
                // Automatically set control width
                label.Width = label.TextWidth;
                this.Width = label.X + label.Width;
            }
        }

        /// <summary>
        /// Gets the button skin interface, allowing skin locations to be
        /// modified.
        /// </summary>
        public ISixSkins ButtonSkin
        {
            get { return button; }
        }

        /// <summary>
        /// Get/Set whether the button is currently checked.
        /// </summary>
        public bool IsChecked
        {
            get { return button.IsChecked; }
            set { button.IsChecked = value; }
        }

        /// <summary>
        /// Gets the checkbox button.
        /// </summary>
        protected ImageButton Button
        {
            get { return button; }
        }
        #endregion

        #region Events
        // Event is simply used to relay clicks on the button
        new public event ClickHandler Click;
        #endregion

        #region Constructors
        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="game">The currently running Game object.</param>
        /// <param name="guiManager">GUIManager that this control is part of.
        /// </param>
        public CheckBox(Game game, GUIManager guiManager)
            : base(game, guiManager)
        {
            #region Create Child Controls
            this.button = new ImageButton(game, guiManager);
            this.label = new Label(game, guiManager);
            #endregion

            #region Add Child Controls
            Add(this.button);
            Add(this.label);
            #endregion

            #region Set Properties
            CanHaveFocus = false;
            #endregion

            #region Set Default Properties
            Width = defaultHeight;
            Height = defaultHeight;
            HMargin = defaultHMargin;
            this.button.SetSkinsFromDefaults(defaultButtonSkin);
            #endregion

            #region Event Handlers
            this.button.Click += new ClickHandler(OnClick);
            #endregion
        }
        #endregion

        /// <summary>
        /// Load default font.
        /// </summary>
        /// <param name="loadAllContent">Which type of content to load.</param>
        protected override void LoadGraphicsContent(bool loadAllContent)
        {
            if (loadAllContent)
                Font = GUIManager.ContentManager.Load<BitmapFont>(defaultFont);

            base.LoadGraphicsContent(loadAllContent);
        }

        /// <summary>
        /// Update controls to respect new margins.
        /// </summary>
        protected void RefreshMargins()
        {
            // Update margin
            label.X = HMargin + button.Width;
            label.Width = Width - button.Width - HMargin;
        }

        #region Event Handlers
        /// <summary>
        /// Update locations of child control.
        /// </summary>
        /// <param name="sender"></param>
        protected override void OnResize(UIComponent sender)
        {
            base.OnResize(sender);

            // Update child sizes
            button.Width = Height;
            button.Height = Height;
            label.Height = Height;

            RefreshMargins();
        }

        /// <summary>
        /// This event handler is called when the button is clicked. It simply
        /// relays the event to the event handler of this control.
        /// </summary>
        /// <param name="sender">Clicked control.</param>
        public void OnClick(UIComponent sender)
        {
            if (Click != null)
                Click.Invoke(this);
        }
        #endregion
    }
}