#region File Description
//-----------------------------------------------------------------------------
// File:      Label.cs
// Namespace: WindowSystem
// Author:    Aaron MacDougall
//-----------------------------------------------------------------------------
#endregion

#region Using Statements
using System;
using System.Diagnostics;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using XNAExtras;
#endregion

namespace XNAWindowSystem
{
    /// <summary>
    /// A graphical text label.
    /// </summary>
    public class Label : DrawableUIComponent
    {
        #region Default Properties
        private static string defaultFont = "Content/Fonts/DefaultFont";
        private static Color defaultColor = Color.Black;

        /// <summary>
        /// Sets the default font.
        /// </summary>
        /// <value>Must be a non-empty string.</value>
        public static string DefaultFont
        {
            set
            {
                Debug.Assert(value != null);
                Debug.Assert(value.Length > 0);
                defaultFont = value;
            }
        }

        /// <summary>
        /// Sets the default text colour.
        /// </summary>
        public static Color DefaultColor
        {
            set { defaultColor = value; }
        }
        #endregion

        #region Fields
        private string text;
        private char cursor;
        private bool isCursorShown;
        private BitmapFont font;
        private Color color;
        #endregion

        #region Properties
        /// <summary>
        /// Get/Set the label text.
        /// </summary>
        /// <value>Must not be null.</value>
        public string Text
        {
            get { return text; }
            set
            {
                Debug.Assert(value != null);

                this.text = value;
                Redraw();
            }
        }

        /// <summary>
        /// Get/Set whether the cursor should be shown.
        /// </summary>
        public bool IsCursorShown
        {
            get { return isCursorShown; }
            set
            {
                if (value != this.isCursorShown)
                {
                    this.isCursorShown = value;
                    Redraw();
                }
            }
        }

        /// <summary>
        /// Sets the text font.
        /// </summary>
        /// <value>Must not be null.</value>
        public BitmapFont Font
        {
            set
            {
                Debug.Assert(value != null);
                this.font = value;
                Redraw();
            }
        }

        /// <summary>
        /// Get/Set the text colour.
        /// </summary>
        public Color Color
        {
            get { return this.color; }
            set
            {
                this.color = value;
                Redraw();
            }
        }

        /// <summary>
        /// Gets the font height.
        /// </summary>
        public int TextHeight
        {
            get
            {
                int result = 0;
                if (this.font != null)
                    result = this.font.LineHeight;
                return result;
            }
        }

        /// <summary>
        /// Gets the width of current text.
        /// </summary>
        public int TextWidth
        {
            get
            {
                int result = 0;
                if (this.font != null)
                    result = this.font.MeasureString(Text);
                return result;
            }
        }
        #endregion

        #region Constructors
        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="game">The currently running Game object.</param>
        /// <param name="guiManager">GUIManager that this control is part of.</param>
        public Label(Game game, GUIManager guiManager)
            : base(game, guiManager)
        {
            this.text = string.Empty;
            this.cursor = '_';
            this.isCursorShown = false;

            #region Properties
            CanHaveFocus = false;
            #endregion

            #region Set Default Properties
            Color = defaultColor;
            #endregion
        }
        #endregion

        /// <summary>
        /// Load default font.
        /// </summary>
        /// <param name="loadAllContent">Which type of content to load.</param>
        protected override void LoadGraphicsContent(bool loadAllContent)
        {
            if (loadAllContent)
                Font = GUIManager.ContentManager.Load<BitmapFont>(defaultFont);

            base.LoadGraphicsContent(loadAllContent);
        }

        /// <summary>
        /// Draws text.
        /// </summary>
        /// <param name="spriteBatch">SpriteBatch to draw control with.</param>
        protected override void DrawControl(SpriteBatch spriteBatch)
        {
            // Draw text
            if (this.font != null)
            {
                string text = "";

                if (this.text != null)
                    text += this.text;

                if (this.isCursorShown)
                    text += this.cursor;

                if (text.Length > 0)
                {
                    this.font.SpriteBatchOverride(spriteBatch);
                    this.font.DrawString(0, 0, this.color, text);
                }
            }
        }

        //protected override void DrawControl(GameTime gameTime, SpriteBatch spriteBatch, float transparency)
        //{
        //    // Draw text
        //    if (this.font != null)
        //    {
        //        string text = "";

        //        if (this.text != null)
        //            text += this.text;

        //        if (this.isCursorShown)
        //            text += this.cursor;

        //        if (text.Length > 0)
        //        {
        //            this.font.SpriteBatchOverride(spriteBatch);
                    
        //            // Fix up alpha
        //            Vector4 vector = this.color.ToVector4();
        //            vector.W = transparency;
        //            Color color = new Color(vector);

        //            this.font.DrawString(AbsolutePosition.X, AbsolutePosition.Y, color, text);
        //        }
        //    }
        //}
    }
}