#region File Description
//-----------------------------------------------------------------------------
// File:      MovableArea.cs
// Namespace: WindowSystem
// Author:    Aaron MacDougall
//-----------------------------------------------------------------------------
#endregion

#region Using Statements
using System;
using Microsoft.Xna.Framework;
using InputEventSystem;
#endregion

namespace XNAWindowSystem
{
    #region Delegates
    /// <summary>
    /// When control begins moving.
    /// </summary>
    /// <param name="sender">Moving control.</param>
    public delegate void StartMovingHandler(UIComponent sender);
    /// <summary>
    /// When a control finishes moving.
    /// </summary>
    /// <param name="sender">Moving control.</param>
    public delegate void EndMovingHandler(UIComponent sender);
    #endregion

    /// <summary>
    /// A non-graphical control that moves it's parent control when dragged
    /// with the mouse.
    /// </summary>
    public class MovableArea : UIComponent
    {
        #region Fields
        private bool isDragging;
        private Point lastLocation;
        #endregion

        #region Events
        public event StartMovingHandler StartMoving;
        public event EndMovingHandler EndMoving;
        #endregion

        #region Constructors
        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="game">The currently running Game object.</param>
        /// <param name="guiManager">GUIManager that this control is part of.</param>
        public MovableArea(Game game, GUIManager guiManager)
            : base(game, guiManager)
        {
            this.isDragging = false;
            this.lastLocation = Point.Zero;
        }
        #endregion

        #region Event Handlers
        /// <summary>
        /// Starts dragging, and invokes the StartMoving event.
        /// </summary>
        /// <param name="args">Mouse event arguments.</param>
        protected override void OnMouseDown(MouseEventArgs args)
        {
            if (args.Button == MouseButtons.Left)
            {
                this.isDragging = true;
                this.lastLocation = args.Position;

                if (StartMoving != null)
                    StartMoving.Invoke(this);
            }
        }

        /// <summary>
        /// Finishes dragging, and invokes the EndMoving event.
        /// </summary>
        /// <param name="args">Mouse event arguments.</param>
        protected override void OnMouseUp(MouseEventArgs args)
        {
            if (args.Button == MouseButtons.Left)
            {
                this.isDragging = false;

                if (EndMoving != null)
                    EndMoving.Invoke(this);
            }
        }

        /// <summary>
        /// Updates parent position if the mouse is being dragged.
        /// </summary>
        /// <param name="args">Mouse event arguments.</param>
        protected override void OnMouseMove(MouseEventArgs args)
        {
            if (this.isDragging)
            {
                int xDiff = args.Position.X - lastLocation.X;
                int yDiff = args.Position.Y - lastLocation.Y;

                if (Parent != null)
                {
                    // By moving parent, all children will automatically update
                    // their position, including this control
                    Parent.X += xDiff;
                    Parent.Y += yDiff;
                }

                this.lastLocation = args.Position;
            }
        }
        #endregion
    }
}