#region File Description
//-----------------------------------------------------------------------------
// File:      RadioGroup.cs
// Namespace: WindowSystem
// Author:    Aaron MacDougall
//-----------------------------------------------------------------------------
#endregion

#region Using Statements
using System;
using System.Diagnostics;
using Microsoft.Xna.Framework;
#endregion

namespace XNAWindowSystem
{
    /// <summary>
    /// Contains multiple RadioButton controls, and handles the exclusively
    /// checked behavior. A radio button on it's own is simply a checkbox with
    /// a different default appearence.
    /// </summary>
    public class RadioGroup : DrawableUIComponent
    {
        #region Fields
        private bool firstButtonClicked;
        #endregion

        #region Constructors
        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="game">The currently running Game object.</param>
        /// <param name="guiManager">GUIManager that this control is part of.</param>
        public RadioGroup(Game game, GUIManager guiManager)
            : base(game, guiManager)
        {
            this.firstButtonClicked = false;
            CanHaveFocus = false;
        }
        #endregion

        /// <summary>
        /// Overridden to prevent any control except RadioButton from being
        /// added.
        /// </summary>
        /// <param name="control">Control to add.</param>
        public override void Add(UIComponent control)
        {
            Debug.Assert(false);
        }

        /// <summary>
        /// Overloaded to prevent any control except RadioButton from being
        /// added.
        /// </summary>
        /// <param name="control">RadioButton to add.</param>
        public void Add(RadioButton control)
        {
            // Set event handler and add control
            control.Click += new ClickHandler(OnClick);
            base.Add(control);
        }

        #region Event Handlers
        /// <summary>
        /// When a child is checked, uncheck all other children.
        /// </summary>
        /// <param name="sender">Clicked RadioButton</param>
        protected void OnClick(UIComponent sender)
        {
            RadioButton checkBox = (RadioButton)sender;

            // Uncheck every other checkbox
            foreach (RadioButton control in Controls)
            {
                if (control != sender)
                    control.IsChecked = false;
            }

            // Ensure that only one checkbox is checked at any one time.
            // Also ensure first click gets throught!
            if (this.firstButtonClicked && !checkBox.IsChecked)
                checkBox.IsChecked = true;
            else
                this.firstButtonClicked = true;
        }
        #endregion
    }
}