#region File Description
//-----------------------------------------------------------------------------
// File:      TextButton.cs
// Namespace: WindowSystem
// Author:    Aaron MacDougall
//-----------------------------------------------------------------------------
#endregion

#region Using Statements
using System;
using System.Diagnostics;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using XNAExtras;
using InputEventSystem;
#endregion

namespace XNAWindowSystem
{
    /// <summary>
    /// A push button comprised of a graphical Bar component, with a text label
    /// on top.
    /// </summary>
    public class TextButton : DrawableUIComponent
    {
        #region Default Properties
        private static int defaultWidth = 65;
        private static int defaultHeight = 25;
        private static int defaultEdgeSize = 12;
        private static string defaultFont = "Content/Fonts/DefaultFont";
        private static DefaultThreeSkins defaultSkin = new DefaultThreeSkins(
            new Rectangle(1, 142, 25, 25),
            new Rectangle(27, 142, 25, 25),
            new Rectangle(52, 142, 25, 25)
            );

        /// <summary>
        /// Sets the default button width.
        /// </summary>
        /// <value>Must be greater than 0.</value>
        public static int DefaultWidth
        {
            set
            {
                Debug.Assert(value > 0);
                DefaultWidth = value;
            }
        }

        /// <summary>
        /// Sets the default button height.
        /// </summary>
        /// <value>Must be greater than 0.</value>
        public static int DefaultHeight
        {
            set
            {
                Debug.Assert(value > 0);
                defaultHeight = value;
            }
        }

        /// <summary>
        /// Sets the default size of left and right sides of button.
        /// </summary>
        /// <value>Must be greater than 0.</value>
        public static int DefaultEdgeSize
        {
            set
            {
                Debug.Assert(value > 0);
                defaultEdgeSize = value;
            }
        }

        /// <summary>
        /// Sets the default button font.
        /// </summary>
        /// <value>Must not be null.</value>
        public static string DefaultFont
        {
            set
            {
                Debug.Assert(value != null);
                defaultFont = value;
            }
        }

        /// <summary>
        /// Gets the default skin locations.
        /// </summary>
        public static IThreeSkins DefaultSkin
        {
            get { return defaultSkin; }
        }
        #endregion

        #region Fields
        private Bar buttonBar;
        private Label label;
        #endregion

        #region Properties
        /// <summary>
        /// Gets the button skin interface, allowing skin locations to be
        /// modified.
        /// </summary>
        public IThreeSkins Skin
        {
            get { return this.buttonBar; }
        }

        /// <summary>
        /// Get/Set size of left and right parts of the button.
        /// </summary>
        /// <value>Must be at least 0.</value>
        public int EdgeSize
        {
            get { return this.buttonBar.EdgeSize; }
            set { this.buttonBar.EdgeSize = value; }
        }

        /// <summary>
        /// Get/Set the button text.
        /// </summary>
        /// <value>Must not be null.</value>
        public string Text
        {
            get { return this.label.Text; }
            set
            {
                this.label.Text = value;
                RefreshLabelPosition();
            }
        }

        /// <summary>
        /// The font used to draw button text.
        /// </summary>
        /// <value>Must not be null.</value>
        public BitmapFont Font
        {
            set
            {
                label.Font = value;
                RefreshLabelPosition();
            }
        }
        #endregion

        #region Constructors
        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="game">The currently running Game object.</param>
        /// <param name="guiManager">GUIManager that this control is part of.</param>
        public TextButton(Game game, GUIManager guiManager)
            : base(game, guiManager)
        {
            #region Create Child Controls
            this.buttonBar = new Bar(game, guiManager);
            this.label = new Label(game, guiManager);
            #endregion

            #region Add Child Controls
            Add(this.buttonBar);
            Add(this.label);
            #endregion

            #region Set Properties
            this.buttonBar.IsVertical = false;
            MinWidth = defaultHeight;
            MinHeight = defaultHeight;
            #endregion

            #region Set Default Properties
            Width = defaultWidth;
            Height = defaultHeight;
            EdgeSize = defaultEdgeSize;
            this.buttonBar.SetSkinsFromDefaults(defaultSkin);
            #endregion
        }
        #endregion

        /// <summary>
        /// Loads the default font.
        /// </summary>
        /// <param name="loadAllContent">Which type of content to load.</param>
        protected override void LoadGraphicsContent(bool loadAllContent)
        {
            if (loadAllContent)
                Font = GUIManager.ContentManager.Load<BitmapFont>(defaultFont);

            base.LoadGraphicsContent(loadAllContent);
        }

        /// <summary>
        /// Sets the label sizes, and centres it on the button.
        /// </summary>
        private void RefreshLabelPosition()
        {
            this.label.Width = this.label.TextWidth;
            this.label.Height = this.label.TextHeight;
            // Centre
            this.label.X = (this.Width - this.label.Width) / 2;
            this.label.Y = (this.Height - this.label.Height) / 2;
        }

        #region Event Handlers
        /// <summary>
        /// Updates button skin to reflect current mouse state.
        /// </summary>
        /// <param name="args">Mouse event arguments.</param>
        protected override void OnMouseOver(MouseEventArgs args)
        {
            base.OnMouseOver(args);

            if (IsPressed)
                this.buttonBar.CurrentSkinState = SkinState.Pressed;
            else
                this.buttonBar.CurrentSkinState = SkinState.Hover;
        }

        /// <summary>
        /// Updates button skin to reflect current mouse state.
        /// </summary>
        /// <param name="args">Mouse event arguments.</param>
        protected override void OnMouseOut(MouseEventArgs args)
        {
            base.OnMouseOut(args);
            this.buttonBar.CurrentSkinState = SkinState.Normal;
        }

        /// <summary>
        /// Updates button skin to reflect current mouse state.
        /// </summary>
        /// <param name="args">Mouse event arguments.</param>
        protected override void OnMouseDown(MouseEventArgs args)
        {
            base.OnMouseDown(args);
            if (args.Button == MouseButtons.Left)
                this.buttonBar.CurrentSkinState = SkinState.Pressed;
        }

        /// <summary>
        /// Updates button skin to reflect current mouse state.
        /// </summary>
        /// <param name="args">Mouse event arguments.</param>
        protected override void OnMouseUp(MouseEventArgs args)
        {
            base.OnMouseUp(args);
            if (args.Button == MouseButtons.Left)
            {
                if (CheckCoordinates(args.Position.X, args.Position.Y))
                    this.buttonBar.CurrentSkinState = SkinState.Hover;
                else
                    this.buttonBar.CurrentSkinState = SkinState.Normal;
            }
        }

        /// <summary>
        /// Refresh child controls when control is resized.
        /// </summary>
        /// <param name="sender">Resized control.</param>
        protected override void OnResize(UIComponent sender)
        {
            base.OnResize(sender);

            if (this.buttonBar != null)
            {
                this.buttonBar.Width = Width;
                this.buttonBar.Height = Height;
            }

            if (this.label != null)
                RefreshLabelPosition();
        }
        #endregion
    }
}